#!/usr/bin/python
'''																	
 				Module including functionality that allows to elaborate					
 					and extract useful information from pcz files.						
'''

import sys
import logging as log

import numpy as np
import math

import pcazip
from MDPlus.analysis.pca import pczfile

def pczdump(args):
    listLong = ['--input', '--evec', '--fluc', '--evals', '--avg', '--coll', '--proj', '--info', '--verbosity',
                '--output', '--rms', '--anim', '--pdb']
    listShort = ['-i', '-e', '-f', '-l', '-a', '-c', '-p', '-n', '-v', '-o', '-r', '-m']

    numArgs = 0
    for i in range(1, len(sys.argv)):
        if sys.argv[i] in listShort and listLong[listShort.index(sys.argv[i])] in sys.argv:
            log.error('''Please, use either the long or the short form of an option but never both! Try again!''')
            sys.exit(-1)

    listUniqOpts = ['--evec', '--fluc', '--evals', '--avg', '--coll', '--proj', '--info', '--rms', '--anim', '-e', '-f',
                    '-l', '-a', '-c', '-p', '-n', '-r', '-m']
    for i in range(1, len(sys.argv)):
        if sys.argv[i] in listUniqOpts:
            numArgs += 1
            if numArgs == 2:
                log.error(
                    '''Please ask for an option at a time from pyPczdump. You have currently been asking for more than one option! Try again!''')
                sys.exit(-1)

    if args.verbosity:
        log.basicConfig(format="%(levelname)s: %(message)s", level=log.DEBUG)
        log.info("Verbose output.")
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")

        # log.info('PCZDUMP analysis in action!')
    if (args.input is None):
        log.error('')
        log.error('''All or any of the mandatory command line arguments is missing. The
		correct usage of pyCoCo should be:''')
        log.error('python ../pczdump.py -i|--input <input-file> [optional arguments]')
        log.error('')
        log.error('''Type "pyPczdump -h" or "pyPczdump --help" for further details.''')
        log.error('')
        sys.exit(-1)

    if args.evec is not None:
        pcz_instance = pczfile.Pczfile(args.input[0])
        e = pcz_instance.evec(int(args.evec))
        if args.output is not None:
            np.savetxt(args.output, np.column_stack((e,)))
        else:
            log.info("Values of eigenvector %d follow:\n" % int(args.evec))
            for i in e:
                print i

    elif args.proj is not None:
        pcz_instance = pczfile.Pczfile(args.input[0])
        p = pcz_instance.proj(int(args.proj))
        if args.output is not None:
            np.savetxt(args.output, np.column_stack((p,)))
        else:
            log.info("Values of the projections of eigenvector %d follow:\n" % int(args.proj))
            for i in p:
                print i
    elif args.evals is True:
        pcz_instance = pczfile.Pczfile(args.input[0])
        evs = pcz_instance.evals()
        if args.output is not None:
            np.savetxt(args.output, np.column_stack((evs,)))
        else:
            log.info("The list of eigenvalues follows:\n")
            for i in evs:
                print i
    elif args.info is True:
        pcz_instance = pczfile.Pczfile(args.input[0])
        if pcz_instance.title is not None:
            log.info("Basic information on the compressed file:\n")
            print "%s\n" % pcz_instance.title
        print "The PCZ file format version is %s\n" % pcz_instance.version
        print "The number of atoms is %d\n" % pcz_instance.natoms
        print "The number of frames is %d.\n" % pcz_instance.nframes
        print "The number of eigenvectors is %d.\n" % pcz_instance.nvecs
        print "The variance captured inside this file is %.2f%%.\n" % (
        100 * pcz_instance.evals().sum() / pcz_instance.quality)
    elif args.fluc is not None:
        pcz_instance = pczfile.Pczfile(args.input[0])
        evec = pcz_instance.evec(int(args.fluc))
        fluc = np.empty((pcz_instance.natoms))
        for i in range(pcz_instance.natoms):
            j = 3 * (i + 1) - 3
            fluc[i] = evec[j] * evec[j] + evec[j + 1] * evec[j + 1] + evec[j + 2] * evec[j + 2]
            fluc[i] = np.sqrt(fluc[i])
        if args.output is not None:
            np.savetxt(args.output, np.column_stack((fluc,)))
        else:
            log.info("Values of the fluctuations associated with eigenvector %d follow:\n" % int(args.fluc))
            for i in fluc:
                print i
    elif args.avg is True:
        pcz_instance = pczfile.Pczfile(args.input[0])
        avg = pcz_instance.avg()

        atom_nr = 0

        if args.pdb is None:
            log.error(
                '''The "--avg" or "-a" option REQUIRES the "--pdb" option. The output average structure will be a pdb file either on disk or on screen! Try again including the "--pdb" option!''')
            sys.exit(-1)
        else:
            if args.output is not None:
                with open(args.output, 'w') as wf:
                    with open(args.pdb) as f:
                        for line in f:
                            if line[:4] == "ATOM":
                                wf.write(line[:30])
                                wf.write('%8.3f%8.3f%8.3f' % (avg[atom_nr][0], avg[atom_nr][1], avg[atom_nr][2]))
                                wf.write(line[54:])
                                atom_nr += 1
                            else:
                                wf.write(line)
                    f.close()
                wf.close()
            else:
                with open(args.pdb) as f:
                    for line in f:
                        if line[:4] == "ATOM":
                            print line[:30] + '{0:8.3f}{1:8.3f}{2:8.3f}'.format(avg[atom_nr][0], avg[atom_nr][1],
                                                                                avg[atom_nr][2]) + line[54:-1]
                            atom_nr += 1
                        else:
                            print line[:-1]
                f.close()

    elif args.coll is True:
        pcz_instance = pczfile.Pczfile(args.input[0])
        col = np.zeros((pcz_instance.nvecs))
        for i in range(pcz_instance.nvecs):
            e = pcz_instance.evec(i)
            for j in range(0, pcz_instance.natoms * 3, 3):
                r2 = e[j] * e[j] + e[j + 1] * e[j + 1] + e[j + 2] * e[j + 2]
                col[i] = col[i] - r2 * math.log(r2)
        col = np.exp(col) / pcz_instance.natoms
        if args.output is None:
            # np.savetxt(args.input[0]+"_collMetric.out",np.column_stack((col,)))
            log.info("Collectivity metric values K for each eigenvector follow.:\n")
            log.info("Modes producing most of the collective motion in the system will have high K values.\n")
            for i in col:
                print i
        else:
            np.savetxt(args.output, np.column_stack((col,)))

    elif args.rms is not None:
        pcz_instance = pczfile.Pczfile(args.input[0])
        if ((int(args.rms) >= 0) and (int(args.rms) < pcz_instance.nframes)):
            s2 = pcz_instance.scores(int(args.rms))
        rmsd = np.zeros((pcz_instance.nframes))
        for i in range(pcz_instance.nframes):
            rmsd[i] = 0.0
            s1 = pcz_instance.scores(i)
            if ((int(args.rms) >= 0) and (int(args.rms) < pcz_instance.nframes)):
                s1 = s1 - s2
            rmsd[i] = np.sum(s1 * s1) / pcz_instance.natoms

        if args.output is not None:
            np.savetxt(args.output, np.column_stack((rmsd,)))
        else:
            for i in rmsd:
                print np.sqrt(i)

    elif args.anim is not None:
        pcz_instance = pczfile.Pczfile(args.input[0])
        avg = pcz_instance.avg()
        evec = pcz_instance.evec(int(args.anim))
        proj = pcz_instance.proj(int(args.anim))

        rmin = np.min(proj)
        rmax = np.max(proj)
        rinc = (rmax - rmin) * 0.1
        # re-create proj...
        proj = np.zeros(20)
        for i in range(1, 6, 1):
            proj[i] = proj[i - 1] + rinc
        for i in range(6, 16, 1):
            proj[i] = proj[i - 1] - rinc
        for i in range(16, 20, 1):
            proj[i] = proj[i - 1] + rinc

        if args.pdb is None:
            log.error(
                '''The "--anim" or "-m" option REQUIRES the "--pdb" option. The output average structure will be a pdb file either on disk or on screen! Try again including the "--pdb" option!''')
            sys.exit(-1)
        else:
            if args.output is not None:
                with open(args.output, 'w') as wf:
                    wf.write('REMARK   Animation of eigenvector %4d\n' % int(args.anim))
                    with open(args.pdb) as f:
                        line = f.readline()
                        prec_curs = 0
                        while (line[:4] != "ATOM"):
                            if line[:5] != "MODEL":
                                wf.write(line)
                            prec_curs = f.tell()
                            line = f.readline()

                        for j in range(20):
                            f.seek(prec_curs, 0)
                            x = avg.reshape((avg.shape[0] * avg.shape[1])) + evec * proj[j]
                            atom_nr = 0
                            wf.write('MODEL  %5d\n' % (j + 1))
                            for line in f:
                                if line[:4] == "ATOM":
                                    wf.write(line[:30])
                                    for m in range(3):
                                        wf.write('%8.3f' % (x[atom_nr + m]))
                                    wf.write(line[54:])
                                    atom_nr += 3
                                elif (line[:3] != "END"):
                                    wf.write(line)
                            wf.write('ENDMDL\n')
                    f.close()
                wf.close()
            else:
                print 'REMARK   Animation of eigenvector {0:4d}'.format(int(args.anim))
                with open(args.pdb) as f:
                    line = f.readline()
                    prec_curs = 0
                    while (line[:4] != "ATOM"):
                        if line[:5] != "MODEL":
                            print line[:-1]
                        prec_curs = f.tell()
                        line = f.readline()

                    for j in range(20):
                        f.seek(prec_curs, 0)
                        x = avg.reshape((avg.shape[0] * avg.shape[1])) + evec * proj[j]
                        atom_nr = 0
                        print 'MODEL  {0:5d}'.format(j + 1)
                        for line in f:
                            if line[:4] == "ATOM":
                                print line[:30] + '{0:8.3f}{1:8.3f}{2:8.3f}'.format(x[atom_nr], x[atom_nr + 1],
                                                                                    x[atom_nr + 2]) + line[54:-1]
                                atom_nr += 3
                            elif (line[:3] != "END"):
                                print line[:-1]
                        print 'ENDMDL'
                f.close()



#!/usr/bin/env python
'''
 The python version of pcazip!!

 Reproduces the functionality of the old fortran- and C-based versions.

 In essence its wraps a simple set of procedures provided by two modules:
       'cofasu' - trajectory file handling
       'pcz'    - PCA analysis

 Stripped down to the bare essentials, the complete procedure is:

 >>> from MDPlus.analysis.pca import pcz
 >>> from MDPlus.core import cofasu
 >>>
 >>> f = cofasu.Fasu('topology.top','trajectory.traj')
 >>> c = cofasu.Cofasu(f)
 >>> p = pcz.Pcz(c)
 >>> p.write('compressed.pcz')

 Everything else is basically analysing and sanity-checking the arguments
 given on the command line.
'''

# General python libraries import.
import os.path as op
import sys
import logging as log

'''
The pcz module provides the core pca capabilities.
'''
import pcazip
from MDPlus.analysis.pca import pcz

'''
The enhanced Universe provided by MDPlus is now used by the cofasu module.
This version does however not use the MDPlus pcz file writer, but uses a
method provided by the pcz module.
from MDPlus import Universe, Writer
'''

import numpy as np
from time import time

'''
Begin by defining the little utility function that parses trajectory
filename strings that have the "trajfile(start:stop:step)" structure.
'''

def input_parse(infile):
    if "(" in infile:
        i = infile.find("(")
        if ")" in infile[i:]:
            j = infile.find(")")
            return infile[:i], infile[i + 1:j]
        else:
            log.error('Malformed trajectory filename: {0}'.format(infile))
            sys.exit(-1)
    else:
        return [infile, ':::']

'''
And now here is a little utility function to convert 'mask' pdb files into
MDAnalysis-style selection strings. Basically it reads the second column
(atom number) and uses it to construct 'bynum' selections. Runs
of consecutive numbers are expressed in start:stop form.
'''

def pdb2selection(pdbfile):
    sel = ''
    i = 0
    j = 0
    with open(pdbfile, 'r') as f:
        for line in f:
            if line.find('ATOM') == 0 or line.find('HETATM') == 0:
                k = int(line.split()[1])
                # the next line catches the initialization process:
                if i == 0:
                    i = k
                    j = k - 1
                # are we in a run of consecutive numbers?:
                if k == j + 1:
                    j = k
                else:
                    # time to write out another selection:
                    sel = sel + ' bynum {0}:{1} or'.format(i, j)
                    i = k
                    j = k
                # end-of-file reached. Make sure last selection is included:
    if i > 0 and j > 0:
        sel = sel + ' bynum {0}:{1} or'.format(i, j)
    if len(sel) > 3:
        # remove the trailing ' or':
        sel = sel[:-3]
    return sel

def jumpfix(sel, **kwargs):
    '''
    Function passed to the Fasu object that provides the possibility
    to fix PBC box jumps. The method is not foolproof, but often works.
    '''
    from MDAnalysis.coordinates.core import triclinic_vectors
    all = sel.selectAtoms('name *')
    cen = sel.selectAtoms(kwargs['centre'])
    shift = triclinic_vectors(cen.dimensions).diagonal()/2 - cen.centerOfGeometry()
#    shift = all.bbox(pbc=True)[1]/2 - cen.centerOfGeometry()
    all.translate(shift)
    all.packIntoBox()
    return all

#############################################################################
#                                                                           #
#                        PCAZIP main function (start)                       #
#                                                                           #
#############################################################################

def pcazip(args):
    # Time the complete run time
    time0start = time()

    if args.mpi:
        from MDPlus.libs import activateMPI

    '''
    The cofasu module provides simple abstractions of complex trajectory data.
    The pcz module provides the core pca capabilities.
    '''
    from MDPlus.core.cofasu import Fasu, Cofasu
    from MDPlus.libs import mpiRelated

    rank = mpiRelated.rank
    size = mpiRelated.size

    if args.verbosity:
        log.basicConfig(format="%(levelname)s: %(message)s", level=log.INFO)
        if rank == 0:
            log.info("Verbose output.")
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")

    if args.tests:
        from subprocess import call
        import os
        try:
            if not op.isdir(op.expanduser('~/test')):
                call(['wget','-P', op.expanduser('~'), 'https://bitbucket.org/ramonbsc/pypcazip/downloads/test.tar.gz'])
                call(['tar', 'xvf', op.expanduser('~/test.tar.gz'), "-C", op.expanduser('~')])
            os.chdir(op.expanduser('~/test'))
            call(['./run_tests.sh'])
            if args.verbosity:
                log.info('Please explore additional testing scripts and related trajectory files as necessary at '+op.expanduser('~/test'))
            sys.exit(0)
        except:
            if (rank == 0) and args.verbosity:
                log.error('')
                log.error('Error while trying to test the correct installation of the pyPcazip suite tools!')
                log.error('')        
            sys.exit(-1)
			
            
    '''
    Input filename and topology filename are mandatory. Hence a check on
    these two parameters should be performed:
    '''
    if (not ((args.input is None) ^ (args.album is None))) or (args.topology is None):
        if rank == 0:
            log.error('')
            log.error('All or any of the mandatory command line arguments is missing. The correct usage is:')
            log.error( 'python ./pcazip.py XOR[(-i|--input <input-file>),(-a|--album) <album-file>] -t|--topology <topology-file> [optional arguments]')
            log.error('')
            log.error('Type "python ./pcazip.py -h" or "python ./pcazip.py --help" for further details.')
            log.error('')
        sys.exit(-1)

    '''
    Multiple album files OR multiple trajectory files are permitted.
    The rule is that all the trajectory files in each album must be
    compatible with a single topology file. If more than one album file
    is specified, then either there should be one topology file that is
    applicable to ALL the album files, or there should be one topology file
    specified for each album file. Similar rules operate in cases where
    multiple trajectory files are specified on the command line: either one
    topology file common to all trajectory files, or one topology file per
    trajectory file, must be given. The same rules operate (independently)
    for the selection and masking options: either there should be one
    selection/mask that applies to ALL trajectory files or albums, or there
    should be one selection/mask for each trajectory file or album.

    Let's check that these rules are being followed. First for albums:
    '''
    if args.input is None:
        na = len(args.album)
        nt = len(args.topology)
        if args.selection is None:
            ns = 1
        else:
            ns = len(args.selection)
        if args.mask is not None:
            ns = max(ns, len(args.mask))
        if nt > 1 and nt != na:
            if rank == 0:
                log.error(("Number of topology files must be one,"
                           " or equal to the number of album files."))
            sys.exit(-1)
        if ns > 1 and ns != na:
            if rank == 0:
                log.error(("Number of masks/selections must be one,"
                           " or equal to the number of album files."))
            sys.exit(-1)
    else:
        # now for trajectories:
        na = len(args.input)
        nt = len(args.topology)
        if args.selection is None:
            ns = 1
        else:
            ns = len(args.selection)
        if args.mask is not None:
            ns = max(ns, len(args.mask))
        if nt > 1 and nt != na:
            if rank == 0:
                log.error(("Number of topology files must be one, or equal"
                           " to the number of trajectory files."))
            sys.exit(-1)
        if ns > 1 and ns != na:
            if rank == 0:
                log.error(("Number of masks/selections must be one, or equal"
                           " to the number of trajectory files."))
            sys.exit(-1)
    '''
        We can now build the key data structures.
        The data structures are:

        uniStr[]:              a list of albums a[], one per topology file.
        a[]:                   a list of trajectory specifiers (each of length 4)
        traj. specifier:       [topfile, trajfile, slice, filter] where trajfile
                               is a string containing the trajectory filename,
                               topfile is the appropriate topology file, slice
                               is a string that defines which snapshots in the
                               trajectory are to be included, using the
                               conventional start:stop:step syntax that e.g.
                               numpy uses to slice arrays, and filter is the atom
                               selection string (MDAnalysis format).

    '''
    uniStr = []
    if args.input is None:
        try:
            # There are one or more album files to process. Within an album,
            # all trajectory files will share the same topology file and filter
            # specification.
            for i in range(len(args.album)):
                log.debug('Reading album file {0}'.format(i))
                # sort out the selection string:
                if args.selection == None:
                    sel = 'name *'
                else:
                    if len(args.selection) == 1:
                        sel = args.selection[0]
                    else:
                        sel = args.selection[i]
                if args.mask is not None:
                    if len(args.mask) == 1:
                        sel = sel + ' and (' + pdb2selection(args.mask[0]) + ')'
                    else:
                        sel = sel + ' and (' + pdb2selection(args.mask[i]) + ')'
                # sort out the topology file string:
                if len(args.topology) == 1:
                    top = args.topology[0]
                else:
                    top = args.topology[i]

                # Files opened in text-mode
                for input_str in open(args.album[i]):
                    # Here, we should figure out whether the input_str contains
                    #  a "\n" and in that case not append a "null"-name file
                    # that would trigger an error in further processing. We
                    # should do this step before calling this function
                    # from the point where we read the single line of the album.
                    if input_str != '\n':
                        # EOLs are converted to '\n'
                        input_str = input_str.rstrip('\n')
                        l = input_parse(input_str)
                        a = [top, l[0], l[1], sel]
                        uniStr.append(a)
        except IOError as e:
            if rank == 0:
                log.error("Problems while tried to process the album file.")
                log.error(("Check whether the album file does exist and whether"
                           " its name matches the name given in input.\n"))
                log.error("I/O error({0}): {1}".format(e.errno, e.strerror))
            sys.exit(-2)
    else:
        '''
        One or more trajectory files have ben specified by the user, rather
        than one or more album files.
        '''
        for i in range(len(args.input)):
            log.debug('Reading trajectory file {0}'.format(i))
            # sort out the selection string:
            if args.selection == None:
                sel = 'name *'
            else:
                if len(args.selection) == 1:
                    sel = args.selection[0]
                else:
                    sel = args.selection[i]
            if args.mask is not None:
                if len(args.mask) == 1:
                    sel = sel + ' and (' + pdb2selection(args.mask[0]) + ')'
                else:
                    sel = sel + ' and (' + pdb2selection(args.mask[i]) + ')'
            # sort out the topology file string:
            if len(args.topology) == 1:
                top = args.topology[0]
            else:
                top = args.topology[i]

            input_str = args.input[i]
            l = input_parse(input_str)
            a = [top, l[0], l[1], sel]
            uniStr.append(a)

    # Now we can create the cofasu:
    f = []
    if args.centre is not None:
        tfunc = jumpfix
        kwargs={'centre':args.centre}
        if rank == 0:
            log.info('Will place group {0} at centre of box to fix jumps'.format(args.centre))
    else:
        tfunc = None
        kwargs = {}
    #
    # To be nice to the user, we check if the netCDF4 module is
    # available, and if not make sure to trap attempts to load
    # AMBER .ncdf format trajectory files.
    #
    try:
        import netCDF4
        nonetCDF4 = False
    except ImportError:
        nonetCDF4 = True

    #Time reading/gathering of the trajectories in parallel
    time1start = time()
    i = 0
    for a in uniStr:

        log.debug('Cofasu:{0} {1} {2} {3}'.format(a[0], a[1], a[2], a[3]))
        if op.splitext(a[1])[1].lower() == '.ncdf' and nonetCDF4:
            log.error('netcdf4-python with the netCDF4 and HDF5 libraries must be installed to read AMBER .ncdf files.\nSee installation instructions at https://code.google.com/p/mdanalysis/wiki/netcdf')
            exit(1)
        f.append(Fasu(a[0], a[1], slice=a[2], filter=a[3], owner=i % size, tfunc=tfunc, **kwargs))
        i += 1
    time1end = time()
    try:
        cf = Cofasu(f)
    except(ValueError):
        if rank == 0:
            log.error('Can\'t compile trajectory files - inconsistent sizes?')
        sys.exit(-1)

    if args.optimise:
        # Replace the cofasu with an optimised copy of iteself
        timeoptstart = time()
        cf = cf.optcopy()
        timeoptend = time()

    if args.trj_output is not None:
        cf.write(args.trj_output)
        log.info('Wrote selected frames and atoms to trajectory file {0}'.format(args.trj_output))

    if args.nopca is False:
        # run the pca analysis:
        if rank == 0:
            log.info('Running pca analysis')
        # Timing the pcz-analysis with trajectories distributed across processors
        time2start = time()
        p = pcz.Pcz(cf, quality=float(args.quality), req_evecs=args.evecs,
        rank=rank, version=args.file_version, preload=(not args.lowmem),
        fastmethod=args.fast)
        time2end = time()
        if rank == 0:
            log.info("Writing compressed trajectory")

        if args.output is not None:
            output_file = args.output
        else:
            # The input trajectory file is a mandatory argument and the check
            # on this has been done previously.
            dir = op.dirname(uniStr[0][0][1])
            base_out_compressed = op.basename(uniStr[0][0][1])
            name_out_compressed = op.splitext(base_out_compressed)[0]
            output_file = op.join(dir, name_out_compressed + "_outputPython.pcz")

        if rank == 0:
            time_write_output_0 = time()
            p.write(output_file)
            time_write_output_1 = time()

        if args.pdb_out is not None:
            cf.writepdb(args.pdb_out, cf.coords(0))
        if rank == 0:
            totTime = time() - time0start
            log.info('Time for appending cofasus: {0:.2f} s, {1:.1f}% total runtime\n'.format(time1end - time1start, (
            time1end - time1start) / totTime * 100))
            if args.optimise:
                log.info('Time for cofasu optimisation: {0:.2f} s, {1:.1f}% total runtime\n'.format(timeoptend - timeoptstart, (timeoptend - timeoptstart) / totTime * 100))
            log.info('Time for pcz-analysis: {0:.2f} s, {1:.1f}% total runtime\n'.format(time2end - time2start, (
            time2end - time2start) / totTime * 100))
            log.info('Time to write the output file: {0:.2f} s, {1:.1f}% total runtime\n'.format(
                time_write_output_1 - time_write_output_0, (time_write_output_1 - time_write_output_0) / totTime * 100))
            log.info('Total run time:: {0:.2f} s\n'.format(totTime))



'''
This routine looks after the MPI-parallelization bits.
It takes care of the case that mpi4py is not available,
or that is it apparently available, but not useable.
'''
import __builtin__
try:
    parallel = parallelbuiltin
except:
    __builtin__.parallelbuiltin = False
    parallel = False

if parallel:
    try:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        size = comm.Get_size()
    except:
        comm = None
        rank = 0
        size = 1
else:
    comm = None
    rank = 0
    size = 1



#!/usr/bin/python

import sys
import os.path as op
import logging as log

import numpy as np

import pcazip
from MDPlus import Universe, Writer
from MDPlus.analysis.pca import pczfile

#######################################################
# MAIN FUNCTION
#######################################################

def pcaunzip(args):
    if args.verbosity:
        log.basicConfig(format="%(levelname)s: %(message)s", level=log.INFO)
        log.info("Verbose output.")
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")

    if ((args.topology is None) or (args.compressed is None) or (args.preload is None)):
        log.error('')
        log.error(
            'All or any of the mandatory command line arguments is missing. The correct usage of PCAUNZIP should be:')
        log.error(
            'python ./pcaunzip.py -t|--topology <topology-file>, -c|--compressed <compressed-file>, -p|--preload <True|False>, [optional arguments]')
        log.error('')
        log.error('Type "python ./pcaunzip.py -h" or "python ./pcaunzip.py --help" for further details.')
        log.error('')
        sys.exit(-1)

    # If no name is provided for the output file, the extension of the npz file is just changed into .dcd.
    if not args.output:
        dir = op.dirname(args.compressed)
        base = op.basename(args.compressed)
        name = op.splitext(base)[0]
        args.output = op.join(dir, name + ".dcd")

    try:
        import netCDF4
        nonetCDF4 = False
    except ImportError:
        nonetCDF4 = True

    if op.splitext(args.output)[1].lower() == '.ncdf' and nonetCDF4:
        log.error('netcdf4-python with the netCDF4 and HDF5 libraries must be installed to read AMBER .ncdf files.\nSee installation instructions at https://code.google.com/p/mdanalysis/wiki/netcdf')
        exit(1)

    # PCAunzip
    log.info("PCAunzipping")
    pfile = pczfile.Pczfile(args.compressed, args.preload)
    w = Writer(args.output, pfile.natoms)

    for ts_index in xrange(pfile.nframes):
        time_step = []
        time_step.append(pfile.frame(ts_index))
        kwargs = {'nparray':np.array(time_step), 'permissive':True} # Need permissive because of a bug in MDAnalysis
        universe = Universe(args.topology, 'whatever.Numpy', **kwargs)
        for ts in universe.trajectory:
            w.write(ts)
    w.close()




__import__('pkg_resources').declare_namespace(__name__)

import sys
import os.path as op
import traceback

from MDAnalysis import Universe as MDAUniverse
from MDAnalysis import Writer
import MDAnalysis

from MDPlus.coordinates import NUMPY

try:
    import mdtraj as md
    mdtraj_available = True
except ImportError:
    mdtraj_available = False

#Register the new readers: 
# NPZ: numpy matrix binary file format, encoding coordinates.
# PCZ: PCAzip & PCAunzip
# NUMPY: (object), for various uses, among them support MDtraj file formats (Binpos, etc)

#MDAnalysis.coordinates._trajectory_readers['NPZ'] = NPZ.NPZReader
#MDAnalysis.coordinates._trajectory_readers_permissive['NPZ'] = NPZ.NPZReader
#MDAnalysis.coordinates._trajectory_writers['NPZ'] = NPZ.NPZWriter

#MDAnalysis.coordinates._trajectory_readers['PCZ'] = PCZ.PCZReader
#MDAnalysis.coordinates._trajectory_readers_permissive['PCZ'] = PCZ.PCZReader
#MDAnalysis.coordinates._trajectory_writers['PCZ'] = PCZ.PCZWriter

MDAnalysis.coordinates._trajectory_readers['NUMPY'] = NUMPY.NUMPYReader
MDAnalysis.coordinates._trajectory_readers_permissive['NUMPY'] = NUMPY.NUMPYReader

# We need to hack the Universe
class Universe(MDAUniverse):

    def __init__(self, topologyfile, coordinatefile=None, **kwargs):

        if not coordinatefile:
            super(Universe,self).__init__(topologyfile, coordinatefile=None, **kwargs)
        
        else:
            format = kwargs.pop('format', None)
            if 'binpos' in op.splitext(coordinatefile)[1].lower():
                format = 'binpos'
            elif 'netcdf' in op.splitext(coordinatefile)[1].lower():
            # Extension *.netcdf may not be automatically recognised by MDAnalysis as "ncdf".
                format = 'ncdf'
            elif format is None:
                format = op.splitext(coordinatefile)[1][1:].lower()
                                   
            try:
                super(Universe, self).__init__(topologyfile, coordinatefile=coordinatefile, format=format, **kwargs)
                
            except TypeError, err:
                if 'binpos' in format:
                    if not mdtraj_available:
                        print "Error: mdtraj module not available. Please, install the BINPOS extra requirements to be able to use this format. The command to do this requirements is: pip install pyPcazip[BINPOS]"
                        sys.exit(0)
                    # Obtain an mdtraj trajectory object from mdtraj.
                    t = md.load_binpos(coordinatefile,topologyfile)
                    # Get coordinates in shape=(n_frames, n_atoms, 3).
                    # MDtraj works in Nanometer Units!! Need to rescale to Angstrom (x10).
                    trajNp = t.xyz * 10 
                    # Re-shape to (n_frames, 3 x n_atoms) where every frame is (a1_x, a1_y, a1_z, a2_x, ... aN_z).
                    trajNp = trajNp.reshape(t.n_frames,-1)
         
                
                    kwargs = {'nparray':trajNp, 'permissive':True} 
                    super(Universe,self).__init__(topologyfile,'x.numpy', **kwargs)
    
                
                else:
                    raise TypeError("Universe.load_new() cannot find an appropriate coordinate reader "
                                "for file %r.\n%r" % (coordinatefile, err))                




#!/usr/bin/python
"""
Routines to parse pcz format files. All routines return numpy
arrays, if appropriate.
"""
import struct
import MDAnalysis.core.qcprot as qcp
import numpy as np
import sys
try:
    import h5py
    h5py_available = True
except ImportError:
    h5py_available = False

class Pczfile:
    def __init__(self, filename, preload=False):
        """
        Initialises a new pcz object with the data from the given
        file. With the optional keyword argument 'preload' set to the
        default value of False, data is extracted directly from the pcz
        file as and when required. If memory is not an issue, setting
        'preload=True' will read all data into private arrays and certain
        subsequent methods calls will be much faster.
        """
        self.filename = filename
        try:
            self.filehandle = open(self.filename, 'rb')
        except IOError as e:
            print "Problems while tried to open a pcz-format file."
            print "I/O error({0}): {1}".format(e.errno, e.strerror)
            return
        self.data = self.filehandle.read(100)
        if self.data[0:4] != 'PCZ4' and self.data[0:4] != 'PCZ6':
                if not h5py_available:
                    print "Error: h5py module not available. Please install the PCZ7 extra requirements to be able to use it. The command to do this requirements is: pip install pyPcazip[PCZ7]"
                    sys.exit(0)
                try:
                    self.filehandle = h5py.File(self.filename, "r")
                    self.data = 'PCZ7'
                except IOError:
                    print 'Error - unrecognised file format: ', self.data[0:4], 'while trying to open a pcz file!'
                    self.filehandle.close()
                    return

        if self.data[0:4] == 'PCZ4' or self.data[0:4] == 'PCZ6':
            self.keydata = struct.unpack('4s80s3if', self.data)
            self.version = self.keydata[0]
            self.title = self.keydata[1]
            self.natoms = self.keydata[2]
            self.nframes = self.keydata[3]
            self.nvecs = self.keydata[4]
            self.quality = self.keydata[5]

            self.data = self.filehandle.read(16)
            self.keydata = struct.unpack('4i', self.data)
            self.extra = self.keydata[3]

            if self.extra == 0:
                self.headerblocksize = 116
            else:
                self.headerblocksize = 116 + 16 * self.natoms

            self.avgblocksize = 3 * self.natoms * 4
            if self.version == 'PCZ4':
                self.evblocksize = 3 * self.natoms * 4 + 4 + self.nframes * 4
            else:
                self.evblocksize = 3 * self.natoms * 4 + 4 + 8 + self.nframes * 2

            self.preloaded = preload
            if preload:
                self.preloaded = False
                self._avg = self.avg()
                self._evals = self.evals()
                self._evecs = self.evecs()
                self._projs = np.zeros((self.nvecs, self.nframes))
                for i in range(self.nvecs):
                    self._projs[i, :] = self.proj(i)
                self.preloaded = True

        elif self.data[0:4] == 'PCZ7':
            self.version = self.filehandle.attrs['version']
            self.title = self.filehandle.attrs['title']
            self.natoms = self.filehandle.attrs['natoms']
            self.nframes = self.filehandle.attrs['nframes']
            self.numframes = self.nframes
            self.n_evecs = self.filehandle.attrs['nvecs']
            self.nvecs = self.filehandle.attrs['nvecs']
            self.quality = self.filehandle.attrs['quality']
            self._evecs = np.array(self.filehandle['evec_dataset'])
            self._evals = np.array(self.filehandle['eval_dataset'])
            self._avg = np.array(self.filehandle['ref_coord_dataset'])

            self.preloaded = preload
            if preload:
                self._projs = np.array(self.filehandle['proj_dataset'])
                self._p0 = np.array(self.filehandle['p0_dataset'])
                self._pinc = np.array(self.filehandle['pinc_dataset'])
        else:
            print 'Error - unrecognised file format: ', self.data[0:4], 'while trying to open a pcz file!'
            return

    def avg(self):
        """
        Returns the average structure contained in the pcz file
        as an (natoms,3) numpy array.
        """
        if self.preloaded:
            return self._avg
        elif self.version == 'PCZ7':
            return self._avg.reshape((self.natoms, 3))
        else:
            self.filehandle.seek(self.headerblocksize)
            self.data = self.filehandle.read(self.avgblocksize)
            return np.array(struct.unpack(str(self.natoms * 3) + 'f',
                                          self.data)).reshape((self.natoms, 3))

    def eval(self, ival):
        """
        Returns an eigenvalue from the file.
        """
        if ival >= self.nvecs:
            print 'Error - only ', self.nvecs, ' eigenvectors present'
            return 0.0
        else:
            if self.preloaded:
                return self._evals[ival]
            elif self.version == 'PCZ7':
                return self._evals[ival]
            else:
                self.filehandle.seek(self.headerblocksize + self.avgblocksize +
                                     ival * self.evblocksize + self.avgblocksize)
                return struct.unpack('f', self.filehandle.read(4))[0]

    def evals(self):
        """
        Returns an array of all eigenvalues in the file.
        """
        if self.preloaded:
            return self._evals
        else:
            evs = np.zeros(self.nvecs)
            for i in range(self.nvecs):
                evs[i] = self.eval(i)
            return evs

    def evec(self, ivec):
        """
        Returns a chosen eigenvector from the file in the
        form of a (3*natoms) numpy array.
        """
        if ivec >= self.nvecs:
            print 'Error - only ', self.nvecs, 'eigenvectors present'
            return None
        else:
            if self.preloaded:
                return self._evecs[ivec, :]
            elif self.version == 'PCZ7':
                return self._evecs[ivec, :]
            else:
                self.filehandle.seek(self.headerblocksize + self.avgblocksize +
                                     ivec * self.evblocksize)
                return np.array(struct.unpack(str(3 * self.natoms) + 'f',
                                              self.filehandle.read(self.avgblocksize)))

    def evecs(self):
        """
        Returns all eigenvectors in the file in the form of a
        (nvecs,3*natoms) numpy array.
        """
        if self.preloaded:
            return self._evecs
        else:
            evs = np.zeros((self.nvecs, self.natoms * 3))
            for i in range(self.nvecs):
                evs[i, :] = self.evec(i)
            return evs

    def proj(self, iproj):
        """
        Returns an array of the projections along a given eigenvector. There
        will be one value per snapshot.
        """
        if iproj >= self.nvecs:
            print 'Error - only ', self.nvecs, 'eigenvectors present'
            return None
        else:
            if self.preloaded:
                if self.version == 'PCZ7':
                    proj_list = []
                    for frame_index in xrange(self.nframes):
                        proj_list.append(self._p0[frame_index] + self._pinc[frame_index] * self._projs[frame_index][iproj])
                return self._projs[iproj, :]

            elif self.version == 'PCZ7':
                proj_list = []
                for frame_index in xrange(self.nframes):
                    p0 = np.array(self.filehandle['p0_dataset'][frame_index])
                    pinc = np.array(self.filehandle['pinc_dataset'][frame_index])
                    i = np.array(self.filehandle['proj_dataset'][frame_index][iproj])
                    proj_list.append(p0 + pinc * i)
                return np.array(proj_list)
            else:
                self.filehandle.seek(self.headerblocksize + self.avgblocksize +
                                     iproj * self.evblocksize + self.avgblocksize + 4)
                if self.version == 'PCZ4':
                    return np.array(struct.unpack(str(self.nframes) + 'f',
                                                  self.filehandle.read(4 * self.nframes)))
                else:
                    ip = list(0 for i in range(self.nframes))
                    p0, pinc = struct.unpack('2f', self.filehandle.read(8))
                    self.filehandle.seek(self.headerblocksize + self.avgblocksize +
                                         iproj * self.evblocksize + self.avgblocksize + 12)
                    ip = struct.unpack(str(self.nframes) + 'h',
                                       self.filehandle.read(self.nframes * 2))
                    return np.array(list(p0 + pinc * i for i in ip))

    def scores(self, framenumber):
        """
        Method that returns the scores (projections) corresponding to
        a chosen snapshot (zero based).
        """
        if framenumber >= self.nframes:
            return None
        elif self.version == 'PCZ7':
            p0 = np.array(self.filehandle['p0_dataset'][framenumber])
            pinc = np.array(self.filehandle['pinc_dataset'][framenumber])
            projs_comp = np.array(self.filehandle['proj_dataset'][framenumber])
            projs = [p0 + pinc * i for i in projs_comp]
            return np.array(projs)
        else:
            x = np.zeros(self.nvecs)
            for i in range(self.nvecs):
                x[i] = self.proj(i)[framenumber]
            return x

    def frame(self, framenumber):
        """
        Method to return the coordinates of the given frame - i.e.
        to decompress a snapshot. The data is returned as a (natoms,3)
        numpy array.
        """
        if (framenumber >= self.nframes):
            return None
        else:
            scores = self.scores(framenumber)
            return self.unmap(scores)

    def unmap(self, scores):
        """
        Method to return the coordinates corresponding to a given
        set of scores. If the scores vector has less than nvec elements,
        Pczfile.expand is called to fill in the missing values.
        """
        if self.version == 'PCZ7':
            projs = np.array(scores).T
            return (np.array((self._avg + np.dot(projs, self._evecs)))).reshape(self.natoms, 3)
        else:
            x = self.avg()
            if len(scores) < self.nvecs:
                scores = self.expand(scores)
            for i in range(self.nvecs):
                x = x + (self.evec(i) * scores[i]).reshape((self.natoms, 3))
            return x

    def expand(self, scores):
        """
        Method to complete a truncated list of scores with values for
        missing eigenvectors. Basically the pcz file is scanned to
        find the frame with the best match to the scores given, the
        missing values are then copied directly from this frame.
        """
        nvecs = self.nvecs
        ns = len(scores)
        if ns >= nvecs:
            return scores
        else:
            closestframe = self.closest(scores)
            newscores = self.scores(closestframe)
            newscores[0:ns] = scores
            return newscores

    def closest(self, scores):
        """
        Method to find the index of the frame with scores closest to the
        target values.
        """
        ns = len(scores)
        if self.preloaded:
            temp = self._projs
        else:
            temp = np.zeros((nvecs, self.nframes))
            for vec in range(nvecs):
                temp[vec, :] = self.proj(vec)

        best = 0
        err = ((temp[0:ns, 0] - scores) * (temp[0:ns, 0] - scores)).sum()
        for frame in range(self.nframes):
            newerr = ((temp[0:ns, frame] - scores) * (temp[0:ns, frame] - scores)).sum()
            if newerr < err:
                err = newerr
                best = frame
        return best

    def map(self, crds):
        """
        Method to map an arbitrary coordinate set onto the PC model. The
        coordinate set should be a (natom,3) array-like object that matches
        (for size) what's in the pczfile. An array of projections will be
        returned, one value for each eignevector in the pcz file.
        """
        crdset = np.array(crds)
        if np.shape(crdset) == (self.natoms, 3):
            avg = np.reshape(self.avg(), (self.natoms, 3))
            rot = np.zeros(9, dtype=np.float64)
            crdset_cog = crdset.sum(axis=0) / crdset.shape[0]
            avg_cog = avg.sum(axis=0) / avg.shape[0]
            c = crdset.copy()
            c -= crdset_cog
            rmsd = qcp.CalcRMSDRotationalMatrix((avg - avg_cog).T.astype(np.float64),
                                                c.T.astype(np.float64),
                                                crdset.shape[0], rot, None)
            R = np.matrix(rot.reshape(3, 3))
            c = c * R
            c += avg_cog
            c -= avg
            prj = np.zeros(self.nvecs)
            for i in range(self.nvecs):
                prj[i] = (np.dot(c.flatten(), self.evec(i)))
            return prj
        else:
            return None




#!/usr/bin/python
'''
Implements the Fasu (Filtered And Sliced Universe) class, and
the Cofasu (Collection of Fasus) class.
'''
import sys
import logging as log
import tempfile
#from os import remove
from time import time
import __builtin__
import os

import numpy as np
from MDAnalysis.analysis.align import *
from MDAnalysis.coordinates.base import Timestep

import MDPlus as mda

try:
    parallel = parallelbuiltin
except:
    __builtin__.parallelbuiltin = False

from MDPlus.libs import mpiRelated

comm = mpiRelated.comm
parallel = mpiRelated.parallel
rank = mpiRelated.rank
size = mpiRelated.size
if parallel:
    from mpi4py import MPI

'''
Two little utility functions. Both work on (N,3) numpy arrays of
coordinates and do least-squares fitting and rmsd calculations
respectively. The USP is that they will not fall over if there are
<3 atoms in a coordinate set.
'''
def aatb(mobile,target,weights=None):
    '''
    Least squares fits the coordinates in mobile to the coordinates in
    target, with optional mass weighting. Both coordinate sets should be
    (N,3) numpy arrays. If weights is given, it should be an array of
    length N. The USP of this method is that it works for N >= 1 whereas
    most standard methods require N >= 3.

    >>> a = np.array([[1,2,3],[3,4,2],[3,4,5],[4,5,7]])
    >>> b = np.array([[2,1,3],[4,2,3],[3,5,5],[4,2,2]])
    >>> fit = np.array(
    ... [[ 1.12516775,  0.79573647,  3.76775837],
    ...  [ 4.09962641,  0.55448804,  4.07499587],
    ...  [ 3.4250583,   3.25851991,  2.9645142 ],
    ...  [ 4.35014754,  5.39125558,  2.19273156]])
    >>> print np.allclose(aatb(a,b),fit)
    True

    >>> w = np.array([1,12,16,1])
    >>> wfit = np.array(
    ... [[ 1.8421418,   0.33714323,  2.23635022],
    ... [ 4.75590912,  1.04010775,  2.36205162],
    ... [ 3.14411401,  3.22518467,  3.6378084 ],
    ... [ 3.25783507,  5.39756436,  4.76378975]])
    >>> print np.allclose(aatb(a,b,weights=w),wfit)
    True

    >>> print aatb(a,a)
    [[ 1.  2.  3.]
     [ 3.  4.  2.]
     [ 3.  4.  5.]
     [ 4.  5.  7.]]

    >>> print aatb(a[0],b[0])
    [2 1 3]

    >>> fit2 = np.array(
    ... [[ 1.65835921,  0.82917961,  3.        ],
    ...  [ 4.34164079,  2.17082039,  3.        ]])
    >>> print np.allclose(aatb(a[0:2],b[0:2]),fit2)
    True

    >>> print aatb(a,b[0:3])
    Traceback (most recent call last):
        ...
    ValueError: coordinate sets not same size

    '''
    if mobile.shape != target.shape:
        raise ValueError("coordinate sets not same size")
    if len(np.atleast_2d(mobile))==1:
# a single point - just place at target:
       fitted = target
    elif len(np.atleast_2d(mobile))==2:
# a vector - align mobile with target:
       midpoint = target.mean(axis=0)
       targetvector = target[1]-target[0]
       mobilevector = mobile[1]-mobile[0]
       tvlength = np.sqrt(np.dot(targetvector,targetvector))
       mvlength = np.sqrt(np.dot(mobilevector,mobilevector))
       scalefactor = mvlength/tvlength
       newvector = targetvector*scalefactor/2
       fitted = np.array([midpoint-newvector,midpoint+newvector])
    else:
# at least three points - use standard methods:
        lmob = mobile-mobile.mean(axis=0)
        targ_com = target.mean(axis=0)
        ltarg = target-targ_com
        R,rms = rotation_matrix(ltarg,lmob,weights=weights)
        fitted = np.dot(lmob, np.array(R))
        fitted = fitted+targ_com
    return fitted

def rmsd(c1,c2,R=False):
    '''
    RMSD between the coordinates in c1 and those in c2.  Both coordinate
    sets should be (N,3) numpy arrays.  The USP of this method is that it
    works for N >= 1 whereas most standard methods require N >= 3. If the
    R argument is set the rotation matrix is also returned.

    >>> a = np.array([[1,2,3],[3,4,2],[3,4,5],[4,5,7]])
    >>> b = np.array([[2,1,3],[4,2,3],[3,5,5],[4,2,2]])
    >>> print np.allclose(rmsd(a,b),2.43251656802)
    True

    >>> print rmsd(a[0],b[0])
    0.0

    >>> print rmsd(a,b[:3])
    Traceback (most recent call last):
        ...
    ValueError: coordinate sets not same size
    '''

    if c1.shape != c2.shape:
        raise ValueError("coordinate sets not same size")

    local1 = c1-c1.mean(axis=0)
    local2 = c2-c2.mean(axis=0)
    if len(np.atleast_2d(c1))==1:
       rms = 0.0
       R=False
    elif len(np.atleast_2d(c1))==2:
        R=False
        fitted1 = aatb(local1,local2)
        diff = fitted1-local2
        rms = np.sqrt((diff*diff).mean())
    else:
        Rot,rms = rotation_matrix(local1,local2)
    if R:
        return rms, Rot
    else:
        return rms

class Fasu:
    def __init__(self, topology, trajectory, filter='name *', slice=None,
                start=0, stop=None, step=1, format=None, owner=0, 
                tfunc=None, **kwargs):
        '''
        Initialises a new fasu object. At its most basic:
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile)

        Missing files raise an IOError:
        >>> f = Fasu('nosuchfile', 'nosuchfile')
        Traceback (most recent call last):
           ...
        IOError: [Errno 2] No such file or directory: 'nosuchfile'

        filter is an (MDAnalysis-style) selection string:
        >>> f = Fasu(topfile, trjfile, filter='name CA')

        slice is a start:stop:step construct:
        >>> f = Fasu(topfile, trjfile, slice='2:14:2')

        If you really want to, you can use start, stop, and step instead.
        BE AWARE: Cofasu uses zero-based indexing!

        format is a hint as to the trajectory file format, required if the
        automatic extension-based recognition rules built into MDAnalysis
        would fail:
        >>> f = Fasu(topfile, '../../test/2ozq.dcd.copy', format='dcd')
        
        The 'owner' argument is for use with MPI. It identifies which
        process 'owns' the Fasu, and so has actual access to the associated
        trajectory file. Fasus created by processes that don't 'own' them
        are not fully initialised until their sync() method is called; this
        may be done any time after their creation, but is also done implicitly
        when they are added to a cofasu so is not something most users will
        need to deal with.

        tfunc allows a user-supplied function to be added, that will
        manipulate coordinates before they are returned to the user through
        the coords() methods. The function must conform to the signature:
        atomSelection = tfunc(atomSelection, **kwargs). One key use is in
        supplying functions that fix PBC issues on the fly.
        kwargs are passed straight to the MDAnalysis Universe. The main use is
        for reading numpy arrays as trajectories, but they also suppy extra
        arguments needed for the optional tfunc.

        '''
        self.topology = topology
        self.trajectory = trajectory
        self.format = format
        self.selection = filter
        self.slice = slice
        self.owner = owner
        self.tfunc = tfunc
        self.kwargs = kwargs

        # check the files are legit.:
        test = open(self.topology, 'r')
        test.close()
        test = open(self.trajectory, 'r')
        test.close()

        if rank == owner:
            self.u = mda.Universe(self.topology,self.trajectory,format=self.format,**kwargs)
            self.totframes = self.u.trajectory.numframes
            self.sel = self.u.selectAtoms(self.selection)
            self.masses = self.sel.masses()
            self.names = self.sel.names()
            self.natoms = len(self.sel)
            self.start = start
            if stop == None:
                self.stop = self.totframes-1
            else:
                self.stop = stop
            self.step = step
            if self.slice is not None:
                self.setslice(slice)
            else:
                self.setstart(start)
                self.setstop(stop)
                self.setstep(step)

        else:
        # initialise with some dummy data
            self.u = None
            self.sel = None
            self.start = start
            self.stop = stop
            self.slice = slice
            self.step = step
            self.totframes = 0
            self.natoms = 0
            self.masses = np.zeros(self.natoms)
            self.names = ['X' for i in range(self.natoms)]

    def sync(self):
        '''
        call this routine after a fasu has been initialised
        to ensure all processes have the correct data
        '''
        if parallel:
            self.start = comm.bcast(self.start,root=self.owner)
            self.stop = comm.bcast(self.stop,root=self.owner)
            self.step = comm.bcast(self.step,root=self.owner)
            self.slice = comm.bcast(self.slice,root=self.owner)
            self.totframes = comm.bcast(self.totframes,root=self.owner)
            self.natoms = comm.bcast(self.natoms,root=self.owner)
            self.massess = comm.bcast(self.masses,root=self.owner)
            self.names = comm.bcast(self.names,root=self.owner)

    def setstart(self,start):
        '''
        Set the first snapshot to be considered, with sanity checking.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile)
        >>> f.setstart(2)

        >>> f.setstart(1000)
        Traceback (most recent call last):
          ...
        ValueError: value for start (1000) exceeds number of frames in the trajectory (25).

        '''
        if start >= self.totframes:
            raise ValueError('value for start ({0}) exceeds number of frames in the trajectory ({1}).'.format(start, self.totframes))
        if (self.stop != None) and (start > self.stop):
            raise ValueError('value for start ({0}) exceeds value for stop ({1}).'.format(start,self.stop))
        self.start = start
        if self.start < 0:
            self.start = 0

    def setstop(self,stop):
        '''
        Set the last snapshot, with sanity checking.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile)
        >>> f.setstop(20)
        
        negative indices are fine:
        >>> f.setstop(-1)

        but trying to go backwards is not:
        >>> f.setstart(10)
        >>> f.setstop(9)
        Traceback (most recent call last):
          ...
        ValueError: value for stop is less than value for start.

        '''
        tmpstop = self.stop

        if stop==None:
            self.stop = self.totframes - 1
        else:
            self.stop = stop

        if self.stop < 0:
            self.stop = self.totframes + self.stop
        if self.stop < self.start:
            self.stop = tmpstop
            raise ValueError('value for stop is less than value for start.')
        if self.stop >= self.totframes:
            self.stop = self.totframes - 1

    def setstep(self,step):
        '''
        Set the stepsize, with sanity checking.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile)
        >>> f.setstep(5)
        
        >>> f.setstep(-2)
        Traceback (most recent call last):
          ...
        ValueError: stepsize must be positive.

        '''
        if step < 1:
            raise ValueError('stepsize must be positive.')
        self.step = step

    def setslice(self,slice):
        '''
        Set the start, stop, and step paremeters based on the slice
        string, which takes the form "start:stop:step". Missing values
        are dealt with in the standard pythonic way.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile)
        >>> f.setslice('5:10:2')
        >>> print f.start, f.stop, f.step
        5 10 2

        >>> f.setslice('::3')
        >>> print f.start, f.stop, f.step
        0 24 3

        >>> f.setslice('5:2')
        Traceback (most recent call last):
          ...
        ValueError: value for stop is less than value for start.

        '''
        sl = slice
        nc = slice.count(':')
        if nc == 0:
            sl = sl+'::'
        elif nc == 1:
            sl = sl+':'

        start = sl.partition(':')[0].partition(':')[0]
        stop = sl.partition(':')[2].partition(':')[0]
        step = sl.partition(':')[2].partition(':')[2]

        if start.isdigit():
            self.setstart(int(start))
        else:
            self.setstart(0)

        if stop.isdigit():
            self.setstop(int(stop))
        else:
            self.setstop(-1)

        if step.isdigit():
            self.setstep(int(step))
        else:
            self.setstep(1)

    def setfilter(self,filter):
        '''
        Applies the filter, in the form of an MDAnalysis selection.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile)
        >>> print f.natoms
        2494
        >>> f.setfilter('backbone')
        >>> print f.natoms
        628
        
        '''
        self.selection = filter
        if rank == self.owner:
            self.sel = self.u.selectAtoms(self.selection)
            self.natoms = len(self.sel)
        else:
            self.sel = None
        if parallel:
            self.natoms = comm.bcast(self.natoms,root=self.owner)

    def numframes(self):
        '''
        Returns the actual number of frames that will be considered.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile)
        >>> print f.numframes()
        25

        >>> f.setslice('5:15:3')
        >>> print f.numframes()
        4

        '''
        n = 1+(self.stop-self.start)/self.step
        return n

    def coords(self,snap):
        '''
        Returns a (natoms,3) numpy array of the selected coordinates
        at the chosen snapshot (zero-based).
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile)
        >>> x0 =np.array([30.35, 62.75, 39.94])
        >>> print np.allclose(f.coords(0)[0], x0)
        True

        >>> xbad = f.coords(1000)
        Traceback (most recent call last):
          ...
        ValueError: snapshot out of range (1000).


        '''
        if rank == self.owner:
            nf = self.numframes()
            if snap < 0:
                snap = nf+snap
            if snap >= nf or snap < 0:
                raise ValueError('snapshot out of range ({0}).'.format(snap))

            snp = self.start + snap * self.step
            if snp < (self.u.trajectory.frame-1):
               self.u.trajectory.rewind()
            while (self.u.trajectory.frame-1) < snp:
               try:
                   self.u.trajectory.next()
               except EOFError:
                   print "EOFError trying to access snapshot ",snap
                   raise EOFError
            if self.tfunc is not None:
                crds = self.tfunc(self.sel,**self.kwargs).ts._pos
            else:
                crds =  self.sel.ts._pos
        else:
            print 'Error - process ',rank,' trying to access fasu owned by ',self.owner
            exit(1)

        return crds


    def fitted_sum(self, target, weights=None):
        '''
        Returns the sum of all the snapshots in the Fasu, as a (N,3)
        numpy array, after they have all been least-squares fitted
        to target. If weights is a list of masses, then weighted
        fitting will be done.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile, filter='name CA')
        >>> target = f.coords(0)
        >>> sum = np.array([783.07746, 1539.3837, 1003.4080])
        >>> print np.allclose(f.fitted_sum(target)[0],sum)
        True

        '''

        if rank == self.owner:
#            print 'process ',rank, 'fitted_sum on: ',self.trajectory
            self.target = target
            sum = np.zeros((self.natoms,3))
            for i in range(self.numframes()):
                mob = self.coords(i)
                sum += aatb(mob,self.target,weights=weights)
        else:
            print 'Error - process ',rank,' trying to access fasu owned by ',self.owner
            exit(1)

        return sum

    def cov(self, target, weights=None, preload=True):
        '''
        Returns the covariance matrix for the trajectory,
        after least-squares fitting to target.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile = '../../test/2ozq.dcd'
        >>> f = Fasu(topfile, trjfile, filter='name CA')
        >>> target = f.coords(0)
        >>> cov = f.cov(target)
        >>> print np.allclose(cov[0,:3], np.array([1.536805, 0.267757, 0.832294]))
        True

        '''

        if rank == self.owner:

            if preload:
                if rank == 0:
                    log.info("Calculating covariance matrix fast method")
                try:
                    trj = np.zeros((self.numframes(), self.natoms*3))
                except MemoryError:
                    print "The input file is too big to be preloaded. Please, try to call pyPcazip with the --nopreload option"
                    sys.exit(0)
                for i in range(self.numframes()):
                    mob = aatb(self.coords(i), target, weights=weights)-target
                    trj[i,:] = mob.flatten()
                covar = np.dot(trj.T, trj.conj())/self.numframes()
            else:

                log.info("Calculating covariance matrix big files method")
                max_ts = 10000 # TODO this should be an argument, and dependant on num atoms
                num_blocks = self.numframes()/max_ts
                ts_out = self.numframes() - (max_ts * num_blocks) # Equivalent to ts_out = self.numframes() % max_ts

                if ts_out != 0:
                    # First block it will process the ts_out snaps witch is the same as the module (self.numframes() % max_ts)
                    covar = np.zeros((self.natoms*3, self.natoms*3))
                    trj = np.zeros((ts_out, self.natoms*3))
                    for j in range(ts_out):
                        mob = aatb(self.coords(j), target, weights=weights)-target
                        mob = mob.flatten()
                        trj[j,:] = mob
                    covar = np.dot(trj.T, trj.conj())/ts_out

                total_ts = ts_out  #Current number of processed frames

                # Loop the rest of blocks, each one of max_ts size in order to contain the amount of used memory
                for i in range(num_blocks):
                    trj = np.zeros((max_ts, self.natoms*3))
                if rank == 0:
                    log.info("Calculating covariance matrix big files method")
                """
                num_vars = self.natoms*3
                aver = np.zeros(num_vars)
                for i in range(self.numframes()):
                    mob = aatb(self.coords(i), target, weights=weights)-target
                    mob = mob.flatten()
                    aver += mob

                # Code to monitor the estimated time of calculus ********************************
                loops = (num_vars*(num_vars+1)/2)*self.numframes()
                """


                max_ts = 1000 # TODO this should be an argument, and dependant on num atoms
                num_blocks = self.numframes()/max_ts
                ts_extra = self.numframes()%max_ts
                covar = np.zeros((self.natoms*3, self.natoms*3))

                # LOOP OVER FULL BLOCKS
                trj = np.zeros((max_ts, self.natoms*3))
                for i in range(num_blocks):
                    for j in range(max_ts):
                        index = total_ts + j
                        mob = aatb(self.coords(index), target, weights=weights)-target
                        mob = mob.flatten()
                        trj[j,:] = mob
                    covar += np.dot(trj.T, trj.conj())

                    # Add the current block_covar to the total covar matrix
                    covar = (covar*total_ts + block_covar*max_ts)/(total_ts+max_ts)
                    total_ts += max_ts

                # REMAINDER BLOCK
                trj = np.zeros((ts_extra, self.natoms*3))
                for j in range(ts_extra):
                    index = num_blocks*max_ts + j
                    mob = aatb(self.coords(index), target, weights=weights)-target
                    mob = mob.flatten()
                    trj[j,:] = mob
                covar += np.dot(trj.T, trj.conj())
                covar = covar/self.numframes()

        else:
            print 'Error - process ', rank, ' trying to access fasu owned by ', self.owner
            exit(1)

        #np.savetxt(self.trajectory+str(preload)+".txt",covar)
        return covar

class Cofasu:
    def __init__(self, fasu, check=None):
        '''
        Initialises a new cofasu object. The argument may be a single fasu
        or a list of them. 
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile1 = '../../test/2ozq.dcd'
        >>> trjfile2 = '../../test/2ozq.xtc'
        >>> f1 = Fasu(topfile, trjfile1, filter='name CA')
        >>> f2 = Fasu(topfile, trjfile2, filter='name CA')
        >>> c1 = Cofasu(f1)

        >>> c2 = Cofasu([f1, f2])

        >>> f2 = Fasu(topfile, trjfile1, filter='name CB')
        >>> c2 = Cofasu([f1, f2])
        Traceback (most recent call last):
          ...
        ValueError: fasu has wrong number of atoms.

        
        The optional 'check' argument is used if more than
        one fasu is to be added to the Cofasu. If check="masses", then in
        addition to the obligatory check that all fasus have the same
        number of atoms, a check will be made that they match, in sequence,
        by atomic mass too. 
        >>> f1 = Fasu(topfile, trjfile1, filter='name CA')
        >>> f2 = Fasu(topfile, trjfile2, filter='name O')
        >>> c2 = Cofasu([f1, f2])

        >>> c2 = Cofasu([f1, f2], check='masses')
        Traceback (most recent call last):
          ...
        ValueError: fasu atom masses don't match.

        If check="names" then the atom names will be
        checked instead.
        >>> f2 = Fasu(topfile, trjfile2, filter='name C')
        >>> c2 = Cofasu([f1, f2], check='masses')

        >>> c2 = Cofasu([f1, f2], check='names')
        Traceback (most recent call last):
          ...
        ValueError: fasu atom names don't match.

        '''
        self.fasulist = []
        if isinstance(fasu,list):
            self.fasulist.append(fasu[0])
            fasu[0].sync()
            self.natoms = fasu[0].natoms
            for i in range(1,len(fasu)):
                self.add(fasu[i], check=check)
        else:
            self.fasulist.append(fasu)
            fasu.sync()
            self.natoms = fasu.natoms

    def __len__(self):
        return len(self.fasulist)

    def __getitem__(self,key):
        return Cofasu(self.fasulist[key])

    def add(self, fasu, check=None):
        '''
        Add a new fasu to the cofasu, if it is congruent (has right number of
        atoms), and, if the check is asked for, that the atomic masses match,
        or the atom names match.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile1 = '../../test/2ozq.dcd'
        >>> trjfile2 = '../../test/2ozq.xtc'
        >>> f1 = Fasu(topfile, trjfile1, filter='name CA')
        >>> f2 = Fasu(topfile, trjfile2, filter='name CA')
        >>> c = Cofasu(f1)
        >>> c.add(f2)
        '''
        fasu.sync()
        if fasu.natoms != self.natoms:
            raise ValueError('fasu has wrong number of atoms.')
        if check == 'names':
            m1 = self.fasulist[0].names
            m2 = fasu.names
            if (m1 == m2).all():
                self.fasulist.append(fasu)
            else:
                raise ValueError("fasu atom names don't match.")
        if check == 'masses':
            m1 = self.fasulist[0].masses
            m2 = fasu.masses
            if (m1 == m2).all():
                self.fasulist.append(fasu)
            else:
                raise ValueError("fasu atom masses don't match.")
        else:
            self.fasulist.append(fasu)

    def numframes(self):
        '''
        Returns the total number of frames in the cofasu
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile1 = '../../test/2ozq.dcd'
        >>> trjfile2 = '../../test/2ozq.xtc'
        >>> f1 = Fasu(topfile, trjfile1, filter='name CA')
        >>> f2 = Fasu(topfile, trjfile2, filter='name CA')
        >>> c = Cofasu(f1)
        >>> c.add(f2)
        >>> print c.numframes()
        50

        '''
        nframes = 0
        for fasu in self.fasulist:
             nframes += fasu.numframes()
        return nframes

    def coords(self,snap,target=None,weights=None):
        '''
        Returns a (natom,3) numpy array of the coordinates of the
        selected snapshot (zero-based). If target is given, then
        the coordinates returned are first least-squares fitted
        to this structure, with optional mass-weighting.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile1 = '../../test/2ozq.dcd'
        >>> trjfile2 = '../../test/2ozq.xtc'
        >>> f1 = Fasu(topfile, trjfile1, filter='name CA')
        >>> f2 = Fasu(topfile, trjfile2, filter='name CA')
        >>> c = Cofasu(f1)
        >>> c.add(f2)
        >>> print np.allclose(c.coords(30)[0], np.array([29.40, 61.36, 40.30]))
        True

        >>> target = c.coords(11)
        >>> print np.allclose(c.coords(30, target)[0], np.array([26.56444, 60.82316, 40.36307]))
        True


        '''
        # first some sanity-checking:
        if snap < 0:
            snap = snap + self.numframes()
        if snap < 0:
            raise ValueError('snapshot out of range.')
        if snap > (self.numframes()-1):
            raise ValueError('snapshot out of range.')
        # now find which Fasu contains the desired snapshot:
        ifas = 0
        nsum = self.fasulist[ifas].numframes()
        while (nsum <= snap) and (ifas < (len(self.fasulist)-1)):
            ifas += 1
            nsum = nsum + self.fasulist[ifas].numframes()
        # now get the desired coordinates:
        nsum = nsum - self.fasulist[ifas].numframes()
        if rank == self.fasulist[ifas].owner:
            crds = self.fasulist[ifas].coords(snap-nsum)
            if target is not None:
                crds = aatb(crds,target,weights=weights)
        else:
            crds = None
        if parallel:
            crds = comm.bcast(crds, root=self.fasulist[ifas].owner)
        return crds

    def fitted_average(self, target=None, error=0.0001, weights=None, maxcyc=10):
        '''
        Calculates the global average structure after converged
        cycles of least-squares fitting. The optional argument
        "target" can specify a coordinate set ((natom,3) numpy array)
        to be used to start the process. Otherwise the first structure
        in the cofasu is used. If weights is a list of masses, then
        weighted fitting will be done.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile1 = '../../test/2ozq.dcd'
        >>> trjfile2 = '../../test/2ozq.xtc'
        >>> f1 = Fasu(topfile, trjfile1, filter='name CA')
        >>> f2 = Fasu(topfile, trjfile2, filter='name CA')
        >>> c = Cofasu(f1)
        >>> c.add(f2)
        >>> print np.allclose(c.fitted_average()[0], np.array([ 31.323149, 61.575379, 40.136298]))
        True

        >>> target = c.coords(0) + 10.0
        >>> print np.allclose(c.fitted_average(target)[0], np.array([ 41.323149, 71.575379, 50.136298]))
        True

        '''
        if target is None:
            self.target = self.coords(0)
        else:
            self.target = target

        err = error + 1.0
        icyc = 0
        while err > error and icyc < maxcyc:
            avg = np.zeros((self.natoms,3))
            totavg = np.zeros((self.natoms,3))
            for f in self.fasulist:
                if rank == f.owner:
                    avg += f.fitted_sum(self.target, weights=weights)

            if parallel:
                totavg = comm.reduce(avg, MPI.SUM, root=0)
                totavg = comm.bcast(totavg, root=0)
            else:
                totavg = avg

            avg = totavg/self.numframes()
            err = rmsd(self.target,avg)
            self.target = avg
            icyc += 1
        return avg

    def cov(self, target=None, weights=None, preload=True):
        '''
        Calculate the covariance matrix for the Cofasu, after least-squares
        fitting of each snapshot to target (if given), with optional weights.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile1 = '../../test/2ozq.dcd'
        >>> trjfile2 = '../../test/2ozq.xtc'
        >>> f1 = Fasu(topfile, trjfile1, filter='name CA')
        >>> f2 = Fasu(topfile, trjfile2, filter='name CA')
        >>> c = Cofasu([f1, f2])
        >>> print np.allclose(c.cov()[0][:3], np.array([ 0.2972165,  0.00554705, 0.17948745]))
        True


        '''
        try:
            covar = np.zeros((self.natoms*3,self.natoms*3))
            totcovar = np.zeros((self.natoms*3,self.natoms*3))
        except MemoryError:
            print "The system contains too many atoms ("+str(self.natoms)+"). Please, try to launch pyPcazip with an atom selection."
            sys.exit(0)

        if target is None:
            time_avg_0 = time()
            target = self.fitted_average()
            time_avg_1 = time()
            if (parallel and (rank == 0)):
                log.info('Cofasu: Time for calculating average structure: {0:.2f} s\n'.format(time_avg_1-time_avg_0))
            else:
                log.info('Cofasu: Time to calculate average structure: {0:.2f} s\n'.format(time_avg_1-time_avg_0))


        time_cov_0 = time()
        for f in self.fasulist:
            if rank == f.owner:
                covar += f.cov(target, weights=weights, preload=preload)*f.numframes()
        if parallel:
            totcovar = comm.reduce(covar, MPI.SUM, root=0)
            totcovar = comm.bcast(totcovar, root=0)
        else:
            totcovar = covar
        covar = totcovar/self.numframes()
        time_cov_1 = time()
        if rank == 0:
            log.info('Cofasu: Time for calculating the covariance matrix: {0:.2f} s\n'.format(time_cov_1-time_cov_0))

        return covar

    def write(self, filename, start=0, end=None):
        '''
        Writes the cofasu out to the specified trajectory file. The
        neccessary topology data is taken from the first fasu - note
        that since not every fasu in a cofasu has to have the same 
        topology, this needs some caution from the user. The optional
        start and end arguments specify the range of frames to write
        (from start to end-1 inclusive).
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile1 = '../../test/2ozq.dcd'
        >>> trjfile2 = '../../test/2ozq.xtc'
        >>> f1 = Fasu(topfile, trjfile1)
        >>> f2 = Fasu(topfile, trjfile2)
        >>> c = Cofasu([f1, f2])
        >>> c.write('/tmp/cofasuwritetest.dcd', end=40)
        >>> f3 = Fasu(topfile, '/tmp/cofasuwritetest.dcd')
        >>> c3 = Cofasu(f3)
        >>> print c3.numframes()
        40

        '''
        if end is None:
            end = self.numframes()
        else:
            end = min(self.numframes(),end)
        if rank == self.fasulist[0].owner:
            w = mda.Writer(filename, self.natoms)
        ts = Timestep(self.natoms)
        for ts_index in xrange(start,end):
            ts._pos=self.coords(ts_index)
            if rank == self.fasulist[0].owner:
                w.write(ts)
        if rank == self.fasulist[0].owner:
            w.close()

    def optcopy(self):
        '''
        Returns an optimised copy of a cofasu. A set of new, dcd format,
        trajectory files are generated, one per MPI process (size). The
        coordinates are then split equally amongst them. A corresponding
        topology file (in pdb format) is also generated. These files are
        then used to generate a new cofasu. Whether or not this is worth
        doing will depend on many factors: the format of the original
        trajectory files, how equal in size they are, how the number of
        files compares to the number of available MPI processes, etc.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile1 = '../../test/2ozq.dcd'
        >>> trjfile2 = '../../test/2ozq.xtc'
        >>> f1 = Fasu(topfile, trjfile1, filter='name CA')
        >>> f2 = Fasu(topfile, trjfile2, filter='name CA')
        >>> c = Cofasu([f1, f2])
        >>> copt = c.optcopy()
        >>> print copt.natoms, copt.numframes()
        157 50

        >>> x1 = c.coords(44)
        >>> x2 = copt.coords(44)
        >>> print np.allclose(x1, x2)
        True

        '''
        if rank == 0:
            top_dir = os.path.dirname(self.fasulist[0].topology)
            topfile = top_dir+'/top_{0}.pdb'.format(rank) if len(top_dir) else 'top_{0}.pdb'.format(rank)
        else:
            topfile="dummy.pdb"
        trjlist = []
        for i in range(size):
            if rank == 0:
                trjlist.append(top_dir+'/trj_{0}.dcd'.format(rank)) if len(top_dir) else trjlist.append('trj_{0}.dcd'.format(rank))
            else:
                trjlist.append("dummy.dcd")
        if parallel:
            topfile = comm.bcast(topfile,root=0)
            trjlist = comm.bcast(trjlist,root=0)
        self.writepdb(topfile,self.coords(0))
        
        blocksize = self.numframes()/size
        if self.numframes()%size != 0:
            blocksize += 1
        start = 0
        end = blocksize
        for t in trjlist:
            self.write(t,start=start, end=end)
            start += blocksize
            end += blocksize
            end = min(end,self.numframes())

        if rank == 0:
            log.info("Files for optimised copy of cofasu created")

        f = []
        for i in range(size):
            f.append(Fasu(topfile,trjlist[i],owner=i))
        optc = Cofasu(f)
        if rank == 0:
            log.info("Optimised copy of cofasu created")
        return optc

    def writerst(self, filename, crds):
        '''
        Writes the coordinates in crds out in AMBER restart format.
        '''

        if rank == 0:
            pos = crds.flatten()

            with open(filename, 'w') as f:
                f.write('created by writerst()\n')
                f.write('{:-6d}\n'.format(self.natoms))

                nfull = self.natoms/2
                npart = self.natoms%2

                start = 0
                for i in range(nfull):
                    end = start+6
                    f.write('{0[0]:12.7f}{0[1]:12.7f}{0[2]:12.7f}{0[3]:12.7f}{0[4]:12.7f}{0[5]:12.7f}\n'.format(pos[start:end]))
                    start += 6
                for i in range(npart):
                    end = start+3
                    f.write('{0[0]:12.7f}{0[1]:12.7f}{0[2]:12.7f}\n'.format(pos[start:end]))


    def writepdb(self, filename, crds):
        '''
        A naughty little hack to enable an arbitrary array of
        coordinates - that match the cofasu for shape - to be
        written out as a pdb format file, by hijacking one of
        the atomSelections in the cofasu.
        '''
        if rank == self.fasulist[0].owner:
            self.fasulist[0].sel.set_positions(crds)
            self.fasulist[0].sel.write(filename)

    def rmsd(self, frame1, frame2):
        '''
        Calculates the rmsd between two frames in a cofasu.
        >>> topfile = '../../test/2ozq.pdb'
        >>> trjfile1 = '../../test/2ozq.dcd'
        >>> f1 = Fasu(topfile, topfile, filter='name CA')
        >>> f2 = Fasu(topfile, trjfile1, filter='name CA')
        >>> c = Cofasu([f1, f2])
        >>> print c.rmsd(0, 0) < 0.0001
        True

        >>> print np.allclose(c.rmsd(0, 20), 1.9321309)
        True

        '''
        f1 = self.coords(frame1)
        f2 = self.coords(frame2)

        f1 = f1 - f1.mean(axis=0)
        f2 = f2 - f2.mean(axis=0)

        return rmsd(f1,f2)

if __name__ == "__main__":
    import doctest
    doctest.testmod()



from MDAnalysis.analysis.align import *
import numpy as np
from scipy import spatial,cluster
from MDPlus.core.cofasu import aatb, rmsd

class Optimizer():
    '''
    Coco optimizer class. Initialised with a set of trajectory data and, 
    optionally, an ideal reference structure, it then provides a method to 
    optimise approximate structures generated by the Coco method (or any other).
    '''
    def __init__(self, cofasu, ideal=None, tol=0.02, log=None):
        '''
        Initialization of the optimizer. A collection of snapshots (in
        the form of a cofasu) are analysed to identify rigid subgroups,
        on the basis of having close-to-invariant interatomic distances.
        The 'ideal' structure provides a source of optimal subgroup
        geometries that will be used in the optimize procedure itself.
        If no 'ideal' structure is provided, one will be chosen from the
        cofasu provided, namely that which has the smallest mean squared 
        error in its interatomic distances from the mean values calculated
        over the whole dataset. The 'tol' parameter sets the threshhold for
        the complete linkage clustering that is used to identify the rigid
        groups. Be aware that if tol is set too tight then rigid groups of
        less than three atoms may be generated, which will (currently) break
        the least-squares fitting procedure that is part of the optimization
        process. 'Log' provides a hook to a logger instance.
        '''
        if log is not None:
            log.info("calculating distances...")
#
# step 1: generate the distance matrix <d>, and the matrix of variances var(d)
#
        d = spatial.distance.pdist(cofasu.coords(0), 'euclidean')
        s = d
        s2 = d*d

        for i in range(1,cofasu.numframes()):
            d = spatial.distance.pdist(cofasu.coords(i),'euclidean')
            s += d
            s2 += d*d

        if log is not None:
            log.info("making k matrix...")
        n = cofasu.numframes()
        self.davg = s/n
        s2 = s2/n
        k = s2-(self.davg*self.davg)
#
# the clustering process. This has two stages. In the first the variance matrix
# is used in a complete-linkage clustering process. In the second stage the
# non-overlapping clusters are expanded to include all singly-linked neigbours.
#
        if log is not None:
            log.info("clustering...")
        z = cluster.hierarchy.complete(k)
        rb_id = cluster.hierarchy.fcluster(z,tol,criterion='distance')

        if log is not None:
            log.info("creating rigid groups...")
        numrb = rb_id.max()
        rbsize = np.zeros(numrb)
        for i in rb_id:
            rbsize[i-1] += 1

        rbmat = np.zeros((numrb,cofasu.natoms))
        sqk = spatial.distance.squareform(k)
#
# This is the srigid groups expansion process:
#
        self.rblist = []
        for i in range(numrb):
            rbatomlist = []
            for j in range(cofasu.natoms):
                if rb_id[j] == i+1:
                    rbatomlist.append(j)
            for j in rbatomlist:
                for l in range(cofasu.natoms):
                    if sqk[j,l] < tol:
                        rbmat[i,l] += 1

            for j in range(cofasu.natoms):
                if rbmat[i,j] >= rbsize[i]:
                    rbmat[i,j] = 1
                else:
                    rbmat[i,j] = 0

            rbatomlist = []
            for j in range(cofasu.natoms):
                if rbmat[i,j] == 1:
                    rbatomlist.append(j)
            self.rblist.append(rbatomlist)

        rbsize = rbmat.sum(axis=1)
        self.dups = rbmat.sum(axis=0)

        if log is not None:
            log.info("number of groups: {0}".format(len(self.rblist)))
#
# Here we convert the variance matrix into a force constant matrix. We
# take care to replace zero values with something small (tol*0.1) before
# inverting it. We also set to zero force constants associated with
# interatomic distances that had a varaince greater than tol.
#
        for i in range(len(k)):
            k[i] = max(k[i],tol*0.1)

        self.k = 1.0/k
        invtol = 1.0/tol
        for i in range(len(self.k)):
            if self.k[i] < invtol:
                self.k[i]=0.0
#
# Here we deal with the ideal structure - selecting one if none was given.
#
        if ideal is not None:
            self.ideal = ideal
            d = spatial.distance.pdist(ideal, 'euclidean')
            e = d-self.davg
#            emin = (e*e*self.k).sum()
            emin = (e*e).mean()
        else:
            d = spatial.distance.pdist(cofasu.coords(0), 'euclidean')
            e = d-self.davg
#            emin = (e*e*self.k).sum()
            emin = (e*e).mean()
            imin = 0
            for i in range(1,cofasu.numframes()):
                d = spatial.distance.pdist(cofasu.coords(i), 'euclidean')
                e = d-self.davg
#                e = (e*e*self.k).sum()
                e = (e*e).mean()
                if e < emin:
                    imin = i
                    emin = e
            self.ideal = cofasu.coords(i)

        if log is not None:
            log.info("ideal structure has energy: {0}".format(emin))
        return

    def optimize(self, coords, dtol=0.1, maxcycles=100):
        '''
        Optimize a crude structure, provided as a (natoms,3) numpy array.
        dtol and maxcycles provide criterea to terminate the iterative 
        optimisation. Dtol is the rmsd betwen structures from succesive
        iterations, maxcycles is the maximum number of iterations.
        ''' 
#
# The method works as follows: Using the coordinatates from the 'ideal'
# structure, each rigid subgroup is sequentially least-squares fitted to
# the target structure. The new coordinates are accumulated. At the end, most
# atoms will have been fitted a number of times, as they will have featured
# in a number of the (often overlapping) rigid fragments. Therefore mean
# coordinates are calculated.
# The rmsd of this new structure is compared to the original, and depending
# on the stopping criterea the process may be iterated or terminated.
#
        opt = coords.copy()
        testopt = np.zeros_like(opt)

        for j in range(maxcycles):
            testopt[:,:] = 0.0
            for i in range(len(self.rblist)):
                testopt[self.rblist[i],:] += aatb(self.ideal.take(self.rblist[i],axis=0),opt.take(self.rblist[i],axis=0))

            testopt[:,0] = testopt[:,0]/self.dups
            testopt[:,1] = testopt[:,1]/self.dups
            testopt[:,2] = testopt[:,2]/self.dups

            r = rmsd(opt,testopt)
            opt = testopt.copy()
            if r < dtol:
                break

        return opt

    def energy(self, coords):
        '''
        calculate the energy of a given structure
        '''
        d = spatial.distance.pdist(coords, 'euclidean')
        e = d-self.davg
#        e = (e*e*self.k).sum()
        e = (e*e).mean()
        return e



#!/usr/bin/env python -W ignore
'''
                 *** The command line interface for pyPczclust ***                                                                                                                                  
'''

import logging as log

import numpy as np
from scipy import ndimage

import pcazip
from MDPlus.analysis.pca import pczfile

def watclust(projs, resolution=10):
    '''
    Watershed clustering. Given a set of coordinate data projs(N,D) where
    N is the number of points and D is the number of dimensions, returns
    a list of labels, of length N, identifying the cluster to which each
    point belongs. "resolution" specifies the number of bins for the
    underlying process of histogramming, and may be a single number or a
    list of length D.
    '''
    ndim = projs.shape[1]
    resol = np.zeros(ndim)
    resol[:] = resolution
    # set the boundaries, r, for the histogram. Include a buffer, so 
    # that there is a clear 1-bin boundary around the map.
    r = []
    min = projs.min(axis=0)
    max = projs.max(axis=0)
    for i in range(ndim):
        buff = (max[i]-min[i])/(resol[i]-2)*1.01
        r.append((min[i]-buff, max[i]+buff))
    # create the histogram
    H, edges = np.histogramdd(projs, bins=resol, range=r)

    # define the function that will locate local maxima in the distibution: 
    def f(arr):
        if np.argmax(arr) == len(arr)/2:
            return 1
        else:
            return 0
    # now find them (calling them 'labels' but at the moment they are
    # really 'maxima', will become labels properly next):
    labels = ndimage.filters.generic_filter(H,f,size=3,mode='constant')
    # now turn them into proper labels:
    ndimage.measurements.label(labels,output=labels)
    # now bump up H to name room in the low-order digits for the label
    maxval = np.max(labels)
    log.info("pyPczclust: {0} clusters found".format(int(maxval)))
    scale = int(10**np.ceil(np.log10(maxval)))
    H = H * scale + labels
    # Define the function that assigns bins to labels. For each currently
    # unlabelled bin, we search around to find the neighbour bin with
    # the highest occupancy. If this bin is already labelled, the current
    # bin gets this label too.
    def f2(arr,scale):
        cval = arr[len(arr)/2]
        if cval > 0 and cval%scale == 0:
            maxval = np.max(arr)
            if maxval%scale > 0:
                cval = cval +maxval%scale
        return cval

    # now we can do the watershed-like clustering. The process is iterative,
    # labels "spread out" in a downhill direction from the maxima until all
    # bins are labelled.
    Hnew = ndimage.filters.generic_filter(H,f2,size=3,mode='constant',extra_arguments=(scale,))
    while np.any(Hnew != H):
        H = Hnew
        Hnew = ndimage.filters.generic_filter(H,f2,size=3,mode='constant',extra_arguments=(scale,))

    H = H%scale
    # now we can go through the list of points and assign a cluster label
    # to each:
    tmplist = []
    for proj in projs:
        indx = []
        i = 0
        for c in proj:
            indx.append(np.digitize((c,c),edges[i])[0]-1)
            i += 1
        tmplist.append(H[tuple(indx)])
    # now reassign labels so largest group has label=1, etc.

    newlab = np.argsort(np.argsort(np.bincount(tmplist)))
    newlab = abs(newlab-newlab.max()) + 1

    out = []
    for i in tmplist:
        out.append(newlab[int(i)])

    # now mark "root" structures with negative indices:

    j =0
    for proj in projs:
        indx = []
        i = 0
        for c in proj:
            indx.append(np.digitize((c,c),edges[i])[0]-1)
            i += 1
        if labels[tuple(indx)] > 0:
            out[j] = -out[j]
        j += 1
    return out

def pczclust(args): 
    '''
    Performs histogram/watershed based clustering on data from a .pcz
    file.
    '''

    if args.verbosity:
        log.basicConfig(format="%(levelname)s: %(message)s", level=log.DEBUG)
        log.info("Verbose output")
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")

    p = pczfile.Pczfile(args.pczfile)

    projs = np.zeros((p.nframes,args.dims))
    for i in range(args.dims):
        projs[:,i] = p.proj(i)

    out = watclust(projs, resolution=args.bins)
    np.savetxt(args.outfile,np.c_[projs,out], fmt=("%8.3f"*args.dims + "%5d"))
